from __future__ import annotations
from sqloquent import SqlModel, RelatedModel, RelatedCollection
from sqloquent.errors import tert, vert
from .ArchivedTransaction import ArchivedTransaction
from .Ledger import Ledger
from .Transaction import Transaction
from .Entry import Entry, EntryType
from time import time
import packify


_empty_dict = packify.pack({})


class Statement(SqlModel):
    connection_info: str = ''
    table: str = 'statements'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'height', 'tx_ids', 'ledger_id', 'balances', 'timestamp',
        'details', 'description',
    )
    id: str
    height: int
    tx_ids: str
    ledger_id: str
    balances: bytes
    timestamp: str
    details: bytes
    description: str|None
    ledger: RelatedModel
    transactions: RelatedCollection
    archived_transactions: RelatedCollection

    @property
    def tx_ids(self) -> list[str]:
        """A list of transaction IDs."""
        return self.data.get('tx_ids', '').split(',')
    @tx_ids.setter
    def tx_ids(self, val: list[str]):
        if type(val) is not list:
            return
        # sort the ids, join into a comma-separated string
        val.sort()
        self.data['tx_ids'] = ','.join(val)

    @property
    def balances(self) -> dict[str, tuple[EntryType, int]]:
        """A dict mapping account IDs to tuple[EntryType, int] balances."""
        balances: dict = packify.unpack(self.data.get('balances', _empty_dict))
        return {
            k: (EntryType(v[0]), v[1])
            for k, v in balances.items()
        }
    @balances.setter
    def balances(self, val: dict[str, tuple[EntryType, int]]):
        tert(type(val) is dict, 'balances must be a dict')
        tert(all([
                type(k) is str and
                type(v) is tuple and
                len(v) == 2 and
                type(v[0]) is EntryType and
                type(v[1]) is int
                for k, v in val.items()
            ]),
            'balances must be a dict of str: tuple[EntryType, int]')
        val = {
            k: (v[0].value, v[1])
            for k, v in val.items()
        }
        self.data['balances'] = packify.pack(val)

    @classmethod
    def calculate_balances(
        cls, txns: list[Transaction|ArchivedTransaction],
        parent_balances: dict[str, tuple[EntryType, int]]|None = None,
        reload: bool = False
    ) -> dict[str, tuple[EntryType, int]]:
        """Calculates the account balances for a list of rolled-up
            transactions. If parent_balances is provided, those are the
            starting balances to which the balances of the rolled-up
            transactions are added. If reload is True, the entries are
            reloaded from the database.
        """
        balances = (parent_balances or {}).copy()
        for txn in txns:
            if reload:
                txn.entries().reload()
            for e in txn.entries:
                e: Entry
                bal = {EntryType.CREDIT: 0, EntryType.DEBIT: 0}
                if e.account_id in balances:
                    bal[balances[e.account_id][0]] = balances[e.account_id][1]
                bal[e.type] += e.amount
                net_credit = bal[EntryType.CREDIT] - bal[EntryType.DEBIT]
                if net_credit >= 0:
                    balances[e.account_id] = (EntryType.CREDIT, net_credit)
                else:
                    balances[e.account_id] = (EntryType.DEBIT, -net_credit)
        return balances

    @classmethod
    def prepare(
            cls, txns: list[Transaction|ArchivedTransaction],
            ledger: Ledger|None = None
        ) -> Statement:
        """Prepare a statement by checking that all txns are for the
            same ledger and summarizing the net account balance changes
            from the transactions and the previous Statement. Raises
            TypeError if there are no txns and no ledger, or if the
            transactions are not all Transaction or ArchivedTransaction
            instances. Raises ValueError if the transactions are not all
            for the same ledger.
        """
        tert(len(txns) > 0 or ledger is not None,
            'must provide either txns or ledger')
        tert(all([
                isinstance(txn, (Transaction, ArchivedTransaction))
                for txn in txns
            ]),
             'all txns must be Transaction or ArchivedTransaction instances'
        )
        if ledger is None:
            ledger = txns[0].ledger
        vert(all([ledger.id == txn.ledger_ids for txn in txns]),
             'all txns must be for the same ledger')

        balances = {}
        height = 0

        # check for parent
        parent: Statement|None = cls.query().equal(
            'ledger_id', ledger.id
        ).order_by('height', direction='desc').first()
        if parent is None:
            balances = cls.calculate_balances(txns)
        else:
            balances = cls.calculate_balances(txns, parent.balances)
            height = parent.height + 1

        # create statement
        stmt = cls({
            'ledger_id': ledger.id,
            'timestamp': str(int(time())),
            'height': height
        })
        stmt.balances = balances
        stmt.tx_ids = [txn.id for txn in txns]
        return stmt

    def validate(self, reload: bool = False) -> bool:
        """Validates that the balances are correct, and that the height
            is 1 + the height of the most recentStatement (if one
            exists).
        """
        if reload:
            self.reload()
            self.transactions().reload()
            self.archived_transactions().reload()

        # check that the balances are correct
        txns = {
            txn.id: txn
            for txn in self.transactions
            if txn.id is not None
        }
        txns.update({
            txn.id: txn
            for txn in self.archived_transactions
            if txn.id is not None
        })
        balances = self.calculate_balances(list(txns.values()))
        if balances != self.balances:
            return False

        # check that the height is correct
        parent: Statement|None = self.query().equal(
            'ledger_id', self.ledger_id
        ).order_by('height', direction='desc').not_equal('id', self.id).first()
        if parent is None:
            return self.height == 0
        return self.height == parent.height + 1

    def trim(self, archive: bool = True) -> None:
        """Trims the transactions and entries summarized in this
            Statement. Returns the number of transactions trimmed. If
            archive is True, the transactions and entries are archived
            before being deleted. Raises ValueError if the Statement is
            not valid.
        """
        vert(self.validate(), 'Statement is not valid')
        self.transactions().reload()
        txns = self.transactions
        for txn in txns:
            txn: Transaction
            txn.entries().reload()
            for e in txn.entries:
                e: Entry
                if archive:
                    e.archive()
                e.delete()
            if archive:
                txn.archive()
            txn.delete()
        return len(txns)



from __future__ import annotations
from sqloquent.asyncql import (
    AsyncSqlModel, AsyncRelatedModel, AsyncRelatedCollection, AsyncQueryBuilderProtocol
)
from .EntryType import EntryType
import packify


_None = packify.pack(None)


class ArchivedEntry(AsyncSqlModel):
    connection_info: str = ''
    table: str = 'archived_entries'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'type', 'amount', 'nonce', 'account_id', 'details', 'description',
        'timestamp',
    )
    id: str
    type: str
    amount: int
    nonce: bytes
    account_id: str
    details: bytes
    description: str|None
    timestamp: str|None
    account: AsyncRelatedModel
    transactions: AsyncRelatedCollection

    def __hash__(self) -> int:
        data = self.encode_value(self._encode(self.data))
        return hash(bytes(data, 'utf-8'))

    # override automatic properties
    @property
    def type(self) -> EntryType:
        """The EntryType of the Entry."""
        return EntryType(self.data['type'])
    @type.setter
    def type(self, val: EntryType):
        if type(val) is not EntryType:
            return
        self.data['type'] = val.value

    @property
    def details(self) -> packify.SerializableType:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', _None))
    @details.setter
    def details(self, val: packify.SerializableType):
        self.data['details'] = packify.pack(val)

    @staticmethod
    def _encode(data: dict|None) -> dict|None:
        if type(data) is not dict:
            return data
        if type(data.get('type', None)) is EntryType:
            data['type'] = data['type'].value
        if type(data.get('details', {})) is not bytes:
            data['details'] = packify.pack(data.get('details', None))
        return data

    @staticmethod
    def _parse(data: dict|None) -> dict|None:
        if type(data) is dict and type(data['amount']) is str:
            data['amount'] = int(data['amount'])
        return data

    @classmethod
    async def insert(cls, data: dict, /, *, suppress_events: bool = False,
            parallel_events: bool = False,
        ) -> ArchivedEntry | None:
        """Ensure data is encoded before inserting."""
        result = await super().insert(
            cls._encode(data), suppress_events=suppress_events,
            parallel_events=parallel_events,
        )
        return result

    @classmethod
    async def insert_many(cls, items: list[dict], /, *, suppress_events: bool = False,
            parallel_events: bool = False,
        ) -> int:
        """Ensure data is encoded before inserting."""
        items = [cls._encode(data) for data in items]
        return await super().insert_many(
            items, suppress_events=suppress_events,
            parallel_events=parallel_events,
        )

    @classmethod
    def query(cls, conditions: dict = None) -> AsyncQueryBuilderProtocol:
        """Ensure conditions are encoded properly before querying."""
        return super().query(cls._encode(conditions))



from context import models, simplebooks
from genericpath import isfile
from os import getcwd
from sqlite3 import OperationalError
from time import time
import os
import unittest


DB_FILEPATH = f'{getcwd()}/tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'


class TestStatementsE2E(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        simplebooks.set_connection_info(DB_FILEPATH)
        super().setUpClass()

    def setUp(self):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        simplebooks.publish_migrations(MIGRATIONS_PATH)
        simplebooks.automigrate(MIGRATIONS_PATH, DB_FILEPATH)
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().tearDown()

    def make_txn(self, credit_acct: models.Account, debit_acct: models.Account,
                 amount: int, details: dict|None = None) -> models.Transaction:
        txn_nonce = os.urandom(16)
        credit_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': credit_acct.id,
            'amount': amount,
            'nonce': txn_nonce,
        })
        debit_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': debit_acct.id,
            'amount': amount,
            'nonce': txn_nonce,
        })
        txn = models.Transaction.prepare(
            [credit_entry, debit_entry],
            str(time()),
            details=details
        )
        return txn

    def test_e2e(self):
        assert models.Account.query().count() == 0
        assert models.Entry.query().count() == 0

        # setup identity, currency, ledger, and some accounts
        identity = models.Identity.insert({'name': 'Test Man'})
        currency = models.Currency.insert({
            'name': 'US Dollar',
            'prefix_symbol': '$',
            'fx_symbol': 'USD',
            'base': 100,
            'unit_divisions': 1,
        })
        ledger = models.Ledger.insert({
            'name': 'General Ledger',
            'identity_id': identity.id,
            'currency_id': currency.id,
            'type': models.LedgerType.CURRENT,
        })
        asset_acct, liability_acct, equity_acct = ledger.setup_basic_accounts()
        asset_acct.save()
        liability_acct.save()
        equity_acct.save()

        # make sub account
        assert len(liability_acct.children) == 0
        liability_sub_acct = models.Account.insert({
            'name': 'Liability Sub Account',
            'type': models.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'parent_id': liability_acct.id,
        })

        # prepare and save some valid transactions
        txn1 = self.make_txn(equity_acct, asset_acct, 10_000_00)
        assert txn1.validate()
        txn1.save()
        txn2 = self.make_txn(liability_acct, equity_acct, 5_000_00)
        assert txn2.validate()
        txn2.save()
        txn3 = self.make_txn(liability_sub_acct, liability_acct, 2_000_00)
        assert txn3.validate()
        txn3.save()

        # check balances
        assert equity_acct.balance() == 5_000_00, equity_acct.balance()
        assert asset_acct.balance() == 10_000_00, asset_acct.balance()
        assert liability_acct.balance() == 3_000_00, liability_acct.balance()
        assert liability_sub_acct.balance() == 2_000_00, liability_sub_acct.balance()
        balances = ledger.balances(reload=True)
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 5_000_00, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 3_000_00, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 2_000_00, balances[liability_sub_acct.id][0]

        # make a Statement with the transactions
        stmt = models.Statement.prepare([txn1, txn2, txn3], ledger)
        assert stmt.validate()
        stmt.save()

        # trim
        trimmed = stmt.trim()
        assert trimmed == 3, trimmed

        # statement should not delete itself during a trim
        assert models.Statement.find(stmt.id) is not None, 'Statement trimmed itself'

        # check balances without the Statement balances
        assert equity_acct.balance() == 0, equity_acct.balance()
        assert asset_acct.balance() == 0, asset_acct.balance()
        assert liability_acct.balance() == 0, liability_acct.balance()
        assert liability_sub_acct.balance() == 0, liability_sub_acct.balance()

        # check balances with the Statement balances
        assert equity_acct.balance(previous_balances=stmt.balances) == 5_000_00, \
            equity_acct.balance(previous_balances=stmt.balances)
        assert asset_acct.balance(previous_balances=stmt.balances) == 10_000_00, \
            asset_acct.balance(previous_balances=stmt.balances)
        assert liability_acct.balance(previous_balances=stmt.balances) == 3_000_00, \
            liability_acct.balance(previous_balances=stmt.balances)
        assert liability_sub_acct.balance(previous_balances=stmt.balances) == 2_000_00, \
            liability_sub_acct.balance(previous_balances=stmt.balances)

        # check archived transactions
        stmt.archived_transactions().reload()
        assert len(stmt.archived_transactions) == 3, stmt.archived_transactions
        for txn in stmt.archived_transactions:
            txn: models.ArchivedTransaction
            assert txn.validate()

        # test empty archived transaction and entry
        (models.ArchivedTransaction()).details
        (models.ArchivedEntry()).details


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from sqloquent import SqlModel, RelatedCollection
import packify


_empty_dict = packify.pack({})


class Identity(SqlModel):
    connection_info: str = ''
    table: str = 'identities'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'name', 'details', 'pubkey', 'seed', 'secret_details', 'description',
    )
    id: str
    name: str
    details: bytes
    pubkey: bytes|None
    seed: bytes|None
    secret_details: bytes|None
    description: str|None
    ledgers: RelatedCollection

    # override automatic property
    @property
    def details(self) -> packify.SerializableType:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', _empty_dict))
    @details.setter
    def details(self, val: packify.SerializableType):
        self.data['details'] = packify.pack(val)

    def public(self) -> dict:
        """Return the public data for cloning the Identity."""
        return {
            k:v for k,v in self.data.items()
            if k not in self.columns_excluded_from_hash
        }



from datetime import datetime


def parse_timestamp(timestamp: str) -> int|None:
    """Helper function to automatically parse a timestamp string into a
        Unix epoch timestamp. Returns None if the timestamp is invalid.
        Supports:
        - Unix epoch as integer string (e.g., "1234567890")
        - Unix epoch as float string (e.g., "1234567890.123")
        - ISO 8601 format strings (e.g., "2023-01-01T00:00:00Z")
        - Other datetime formats parseable by datetime.fromisoformat() or datetime.strptime()
    """
    if type(timestamp) is not str:
        raise ValueError('timestamp must be a string')

    if not timestamp:
        return 0

    # try parsing result of str(time())
    try:
        return int(float(timestamp))
    except (ValueError, TypeError):
        pass

    # try ISO 8601 format
    try:
        if 'T' in timestamp or '+' in timestamp or timestamp.endswith('Z'):
            dt = datetime.fromisoformat(timestamp.replace('Z', '+00:00'))
        else:
            # Try parsing without timezone
            dt = datetime.fromisoformat(timestamp)
        return int(dt.timestamp())
    except (ValueError, TypeError):
        pass

    # try some other common formats
    common_formats = [
        '%Y-%m-%d %H:%M:%S',
        '%Y-%m-%d %H:%M:%S.%f',
        '%Y/%m/%d %H:%M:%S',
        '%d/%m/%Y %H:%M:%S',
        '%m/%d/%Y %H:%M:%S',
    ]
    for fmt in common_formats:
        try:
            dt = datetime.strptime(timestamp, fmt)
            return int(dt.timestamp())
        except (ValueError, TypeError):
            continue

    return None



from .Account import Account
from .AccountCategory import AccountCategory
from .AccountType import AccountType
from .ArchivedEntry import ArchivedEntry
from .ArchivedTransaction import ArchivedTransaction
from .Currency import Currency
from .Customer import Customer
from .Entry import Entry
from .EntryType import EntryType
from .Identity import Identity
from .Ledger import Ledger
from .LedgerType import LedgerType
from .Statement import Statement
from .Transaction import Transaction
from .Vendor import Vendor
from sqloquent.asyncql import (
    async_contains, async_within, async_has_many, async_belongs_to,
)


Identity.ledgers = async_has_many(Identity, Ledger, 'identity_id')
Ledger.owner = async_belongs_to(Ledger, Identity, 'identity_id')

Ledger.currency = async_belongs_to(Ledger, Currency, 'currency_id')
Currency.ledgers = async_has_many(Currency, Ledger, 'currency_id')

Ledger.accounts = async_has_many(Ledger, Account, 'ledger_id')
Account.ledger = async_belongs_to(Account, Ledger, 'ledger_id')

Account.children = async_has_many(Account, Account, 'parent_id')
Account.parent = async_belongs_to(Account, Account, 'parent_id')

Account.category = async_belongs_to(Account, AccountCategory, 'category_id')
AccountCategory.accounts = async_has_many(AccountCategory, Account, 'category_id')

Account.entries = async_has_many(Account, Entry, 'account_id')
Entry.account = async_belongs_to(Entry, Account, 'account_id')

Entry.transactions = async_within(Entry, Transaction, 'entry_ids')
Transaction.entries = async_contains(Transaction, Entry, 'entry_ids')

Transaction.ledgers = async_contains(Transaction, Ledger, 'ledger_ids')
Ledger.transactions = async_within(Ledger, Transaction, 'ledger_ids')

Statement.ledger = async_belongs_to(Statement, Ledger, 'ledger_id')
Ledger.statements = async_has_many(Ledger, Statement, 'ledger_id')

Statement.transactions = async_contains(Statement, Transaction, 'tx_ids')
Transaction.statements = async_within(Transaction, Statement, 'tx_ids')

Statement.archived_transactions = async_contains(Statement, ArchivedTransaction, 'tx_ids')
ArchivedTransaction.statements = async_within(ArchivedTransaction, Statement, 'tx_ids')

ArchivedEntry.transactions = async_within(ArchivedEntry, ArchivedTransaction, 'entry_ids')
ArchivedTransaction.entries = async_contains(ArchivedTransaction, ArchivedEntry, 'entry_ids')

ArchivedEntry.account = async_belongs_to(ArchivedEntry, Account, 'account_id')
Account.archived_entries = async_has_many(Account, ArchivedEntry, 'account_id')

ArchivedTransaction.ledgers = async_contains(ArchivedTransaction, Ledger, 'ledger_ids')
Ledger.archived_transactions = async_within(Ledger, ArchivedTransaction, 'ledger_ids')


def set_connection_info(db_file_path: str):
    """Set the connection info for all models to use the specified
        sqlite3 database file path.
    """
    Account.connection_info = db_file_path
    AccountCategory.connection_info = db_file_path
    ArchivedEntry.connection_info = db_file_path
    ArchivedTransaction.connection_info = db_file_path
    Currency.connection_info = db_file_path
    Customer.connection_info = db_file_path
    Entry.connection_info = db_file_path
    Identity.connection_info = db_file_path
    Ledger.connection_info = db_file_path
    Statement.connection_info = db_file_path
    Transaction.connection_info = db_file_path
    Vendor.connection_info = db_file_path


# no longer needed
del async_contains
del async_belongs_to
del async_within
del async_has_many




from __future__ import annotations
from sqloquent.asyncql import AsyncSqlModel, AsyncRelatedCollection
from sqloquent.errors import vert
from .ArchivedEntry import ArchivedEntry, EntryType
import packify


_empty_dict = packify.pack({})


class ArchivedTransaction(AsyncSqlModel):
    connection_info: str = ''
    table: str = 'archived_transactions'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'entry_ids', 'ledger_ids', 'timestamp', 'details', 'description',
    )
    id: str
    entry_ids: str
    ledger_ids: str
    timestamp: str
    details: bytes
    description: str|None
    entries: AsyncRelatedCollection
    ledgers: AsyncRelatedCollection
    statements: AsyncRelatedCollection

    # override automatic properties
    @property
    def details(self) -> dict[str, bytes]:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', _empty_dict))
    @details.setter
    def details(self, val: dict[str, bytes]):
        if type(val) is not dict:
            return
        if not all([type(k) is str and type(v) is bytes for k, v in val.items()]):
            return
        self.data['details'] = packify.pack(val)

    @classmethod
    def _encode(cls, data: dict|None) -> dict|None:
        """Encode values for saving."""
        if type(data) is not dict:
            return None
        if type(data.get('details', {})) is dict:
            data['details'] = packify.pack(data.get('details', {}))
        return data

    async def validate(self, reload: bool = False) -> bool:
        """Determines if a Transaction is valid using the rules of accounting.
            Raises TypeError for invalid arguments. Raises ValueError if the
            entries do not balance for each ledger; or if any of the entries is
            contained within an existing Transaction. If reload is set to True,
            entries and accounts will be reloaded from the database.
        """
        if reload:
            await self.entries().reload()

        # first check that all ledgers balance
        ledgers = {}
        entry: ArchivedEntry
        for entry in self.entries:
            if reload:
                await entry.account().reload()
            if entry.account.ledger_id not in ledgers:
                ledgers[entry.account.ledger_id] = {'Dr': 0, 'Cr': 0}
            if entry.type in (EntryType.CREDIT, EntryType.CREDIT.value):
                ledgers[entry.account.ledger_id]['Cr'] += entry.amount
            else:
                ledgers[entry.account.ledger_id]['Dr'] += entry.amount

        for ledger_id, balances in ledgers.items():
            vert(balances['Cr'] == balances['Dr'],
                f"ledger {ledger_id} unbalanced: {balances['Cr']} Cr != {balances['Dr']} Dr")

        return True

    async def save(self, reload: bool = False) -> ArchivedTransaction:
        """Validate the transaction, save the entries, then save the
            transaction.
        """
        assert await self.validate(reload), 'cannot save an invalid Transaction'
        for e in self.entries:
            await e.save()
        return await super().save()



from asyncio import run
from context import asyncql, simplebooks
from genericpath import isfile
from sqlite3 import OperationalError
from time import time
import os
import unittest


DB_FILEPATH = 'tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'


class TestAsyncBasicE2E(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        asyncql.set_connection_info(DB_FILEPATH)
        super().setUpClass()

    def setUp(self):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().tearDown()

    def automigrate(self):
        simplebooks.publish_migrations(MIGRATIONS_PATH)
        simplebooks.automigrate(MIGRATIONS_PATH, DB_FILEPATH)

    def test_e2e(self):
        with self.assertRaises(OperationalError):
            run(asyncql.Account.query().count())
        self.automigrate()
        assert run(asyncql.Account.query().count()) == 0

        # setup account categories
        equity_acct_cat = run(asyncql.AccountCategory.insert({
            'name': 'Equity',
            'ledger_type': asyncql.LedgerType.CURRENT,
            'destination': 'Balance Sheet',
        }))
        assert equity_acct_cat is not None
        assert run(asyncql.AccountCategory.find(equity_acct_cat.id)) is not None
        asset_acct_cat = run(asyncql.AccountCategory.insert({
            'name': 'Asset',
            'ledger_type': asyncql.LedgerType.CURRENT,
            'destination': 'Balance Sheet',
        }))
        liability_acct_cat = run(asyncql.AccountCategory.insert({
            'name': 'Liability',
            'ledger_type': asyncql.LedgerType.CURRENT,
            'destination': 'Balance Sheet',
        }))

        # setup identity, currency, ledger, and some accounts
        identity = run(asyncql.Identity.insert({'name': 'Test Man'}))
        currency = run(asyncql.Currency.insert({
            'name': 'US Dollar',
            'prefix_symbol': '$',
            'fx_symbol': 'USD',
            'base': 100,
            'unit_divisions': 1,
        }))
        ledger: asyncql.Ledger = run(asyncql.Ledger.insert({
            'name': 'General Ledger',
            'identity_id': identity.id,
            'currency_id': currency.id,
            'type': asyncql.LedgerType.CURRENT,
        }))
        equity_acct = run(asyncql.Account.insert({
            'name': 'General Equity',
            'type': asyncql.AccountType.EQUITY,
            'ledger_id': ledger.id,
            'category_id': equity_acct_cat.id,
        }))
        asset_acct = run(asyncql.Account.insert({
            'name': 'General Asset',
            'type': asyncql.AccountType.ASSET,
            'ledger_id': ledger.id,
            'category_id': asset_acct_cat.id,
        }))
        liability_acct = run(asyncql.Account.insert({
            'name': 'General Liability',
            'type': asyncql.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'category_id': liability_acct_cat.id,
        }))

        # make sub account
        assert len(liability_acct.children) == 0
        liability_sub_acct = run(asyncql.Account.insert({
            'name': 'Liability Sub Account',
            'type': asyncql.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'parent_id': liability_acct.id,
        }))
        assert liability_sub_acct.parent is not None, liability_sub_acct.parent
        assert liability_sub_acct.parent.id == liability_acct.id, liability_sub_acct.parent
        run(liability_acct.children().reload())
        assert len(liability_acct.children) == 1
        assert liability_acct.children[0].id == liability_sub_acct.id

        assert equity_acct.category.id == equity_acct_cat.id
        assert asset_acct.category.id == asset_acct_cat.id
        assert liability_acct.category.id == liability_acct_cat.id

        assert len(liability_acct_cat.accounts) == 1, liability_acct_cat.accounts

        # make a vendor and a customer
        vendor = asyncql.Vendor({'name': 'Vendor-san', 'code': '1530'})
        vendor.details = 'some details'
        vendor.description = 'a vendor'
        run(vendor.save())
        customer = asyncql.Customer({'name': 'Customer-sama', 'code': '2300'})
        customer.details = '1234'
        customer.description = 'a customer'
        run(customer.save())

        # test empty transaction and entry
        (asyncql.Transaction()).details
        (asyncql.Entry()).details

        # prepare and save a valid transaction
        txn_nonce = os.urandom(16)
        equity_entry = asyncql.Entry({
            'type': asyncql.EntryType.CREDIT,
            'account_id': equity_acct.id,
            'amount': 10_000_00,
            'nonce': txn_nonce,
        })
        equity_entry.account = equity_acct
        asset_entry = asyncql.Entry({
            'type': asyncql.EntryType.DEBIT,
            'account_id': asset_acct.id,
            'amount': 10_000_00,
            'nonce': txn_nonce,
        })
        asset_entry.account = asset_acct
        txn = run(asyncql.Transaction.prepare(
            [equity_entry, asset_entry], str(time()),
            details='Starting capital asset'
        ))
        assert run(txn.validate())
        run(equity_entry.save())
        run(asset_entry.save())
        run(txn.save())
        # reload txn from database and validate it
        txn: asyncql.Transaction = run(asyncql.Transaction.find(txn.id))
        assert run(txn.validate(reload=True))
        assert len(asset_entry.transactions) == 1
        assert asset_entry.transactions[0].id == txn.id, asset_entry.transactions

        # check balances
        assert run(equity_acct.balance()) == 10_000_00, run(equity_acct.balance())
        assert run(asset_acct.balance()) == 10_000_00, [run(asset_acct.balance()), asset_acct.entries]
        assert run(liability_acct.balance()) == 0, run(liability_acct.balance())
        balances = run(ledger.balances(True))
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 10_000_00, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 0, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 0, balances[liability_sub_acct.id][0]

        # prepare and save valid transaction for liability sub account
        txn_nonce = os.urandom(16)
        equity_entry = run(asyncql.Entry.insert({
            'type': asyncql.EntryType.DEBIT,
            'account_id': equity_acct.id,
            'amount': 9_99,
            'nonce': txn_nonce,
        }))
        liability_entry = run(asyncql.Entry.insert({
            'type': asyncql.EntryType.CREDIT,
            'account_id': liability_sub_acct.id,
            'amount': 9_99,
            'nonce': txn_nonce,
        }))
        txn = run(asyncql.Transaction.prepare(
            [equity_entry, liability_entry],
            str(time()),
            {
                'vendor_id': vendor.id,
                'customer_id': customer.id,
            }
        ))
        assert run(txn.validate())
        run(txn.save())

        # check balances
        assert run(equity_acct.balance()) == 10_000_00-9_99, run(equity_acct.balance())
        assert run(asset_acct.balance()) == 10_000_00, run(asset_acct.balance())
        assert run(liability_sub_acct.balance()) == 9_99, run(liability_acct.balance())
        assert run(liability_acct.balance()) == 9_99, run(liability_acct.balance())
        assert run(liability_acct.balance(False)) == 0, run(liability_acct.balance(False))
        balances = run(ledger.balances(True))
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 10_000_00-9_99, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 0, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 9_99, balances[liability_sub_acct.id][0]

        # prepare invalid transaction: reused entries
        with self.assertRaises(ValueError) as e:
            txn = run(asyncql.Transaction.prepare([equity_entry, asset_entry], str(int(time()))))
        assert 'already contained within a Transaction' in str(e.exception)

        # prepare invalid transaction: unbalanced entries
        txn_nonce = os.urandom(16)
        equity_entry = asyncql.Entry({
            'type': asyncql.EntryType.CREDIT,
            'account_id': equity_acct.id,
            'amount': 10_00,
            'nonce': txn_nonce,
        })
        equity_entry.account = equity_acct
        asset_entry = asyncql.Entry({
            'type': asyncql.EntryType.DEBIT,
            'account_id': asset_acct.id,
            'amount': 10_01,
            'nonce': txn_nonce,
        })
        asset_entry.account = asset_acct
        with self.assertRaises(ValueError) as e:
            txn = run(asyncql.Transaction.prepare([equity_entry, asset_entry], str(int(time()))))
        assert 'unbalanced' in str(e.exception)

        # test Entry.insert_many
        entries = run(asyncql.Entry.insert_many([
            {
                'type': asyncql.EntryType.CREDIT,
                'account_id': equity_acct.id,
                'amount': 10_000_00,
                'nonce': os.urandom(16),
            },
            {
                'type': asyncql.EntryType.DEBIT,
                'account_id': asset_acct.id,
                'amount': 10_000_00,
                'nonce': os.urandom(16),
            },
        ]))
        assert entries == 2, entries

        # test additional models
        async def test_async_additional_models():
            customer = asyncql.Customer({
                'name': 'John Doe',
                'code': 'JD',
                'description': 'test customer (async)',
            })
            customer.details = {'foo': 'bar'}
            await customer.save()
            assert customer.id is not None
            assert customer.details == {'foo': 'bar'}
            vendor = asyncql.Vendor({
                'name': 'Acme Inc.',
                'code': 'ACME',
                'description': 'test vendor (async)',
            })
            vendor.details = {'foo': 'bar'}
            await vendor.save()
            assert vendor.id is not None
            assert vendor.details == {'foo': 'bar'}

        run(test_async_additional_models())


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from sqloquent.asyncql import AsyncSqlModel, AsyncRelatedCollection
from sqloquent.errors import vert, tert
from .ArchivedTransaction import ArchivedTransaction
from .Entry import Entry, EntryType
import packify


_empty_dict = packify.pack({})


class Transaction(AsyncSqlModel):
    connection_info: str = ''
    table: str = 'transactions'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'entry_ids', 'ledger_ids', 'timestamp', 'details', 'description',
    )
    id: str
    entry_ids: str
    ledger_ids: str
    timestamp: str
    details: bytes
    description: str|None
    entries: AsyncRelatedCollection
    ledgers: AsyncRelatedCollection

    # override automatic properties
    @property
    def details(self) -> dict[str, bytes]:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', _empty_dict))
    @details.setter
    def details(self, val: dict[str, bytes]):
        if type(val) is not dict:
            return
        if not all([type(k) is str and type(v) is bytes for k, v in val.items()]):
            return
        self.data['details'] = packify.pack(val)

    @classmethod
    def _encode(cls, data: dict|None) -> dict|None:
        """Encode values for saving."""
        if type(data) is not dict:
            return None
        if type(data.get('details', {})) is dict:
            data['details'] = packify.pack(data.get('details', {}))
        return data

    @classmethod
    async def prepare(cls, entries: list[Entry], timestamp: str,
                      details: packify.SerializableType = None,
                      reload: bool = False) -> Transaction:
        """Prepare a transaction. Raises TypeError for invalid arguments.
            Raises ValueError if the entries do not balance for each
            ledger; or if any of the entries is contained within an existing
            Transaction. Entries and Transaction will have IDs generated but
            will not be persisted to the database and must be saved separately.
        """
        tert(type(entries) is list and all([type(e) is Entry for e in entries]),
            'entries must be list[Entry]')
        tert(type(timestamp) is str, 'timestamp must be str')

        ledgers = set()
        for entry in entries:
            entry.id = entry.id if entry.id else entry.generate_id()
            if reload:
                await entry.account().reload()
            vert(await Transaction.query().contains('entry_ids', entry.id).count() == 0,
                 f"entry {entry.id} is already contained within a Transaction")
            ledgers.add(entry.account.ledger_id)

        txn = cls({
            'entry_ids': ",".join(sorted([
                e.id if e.id else e.generate_id()
                for e in entries
            ])),
            'ledger_ids': ",".join(sorted(list(ledgers))),
            'timestamp': timestamp,
        })
        txn.details = details
        txn.entries = entries
        assert await txn.validate(reload), \
            'transaction validation failed'
        txn.id = txn.generate_id()
        return txn

    async def validate(self, reload: bool = False) -> bool:
        """Determines if a Transaction is valid using the rules of accounting.
            Raises TypeError for invalid arguments. Raises ValueError if the
            entries do not balance for each ledger; or if any of the entries is
            contained within an existing Transaction. If reload is set to True,
            entries and accounts will be reloaded from the database.
        """
        if reload:
            await self.entries().reload()

        # first check that all ledgers balance
        ledgers = {}
        entry: Entry
        for entry in self.entries:
            if reload:
                await entry.account().reload()
            if entry.account.ledger_id not in ledgers:
                ledgers[entry.account.ledger_id] = {'Dr': 0, 'Cr': 0}
            if entry.type in (EntryType.CREDIT, EntryType.CREDIT.value):
                ledgers[entry.account.ledger_id]['Cr'] += entry.amount
            else:
                ledgers[entry.account.ledger_id]['Dr'] += entry.amount

        for ledger_id, balances in ledgers.items():
            vert(balances['Cr'] == balances['Dr'],
                f"ledger {ledger_id} unbalanced: {balances['Cr']} Cr != {balances['Dr']} Dr")

        return True

    async def save(self, reload: bool = False) -> Transaction:
        """Validate the transaction, save the entries, then save the
            transaction.
        """
        assert await self.validate(reload), 'cannot save an invalid Transaction'
        for e in self.entries:
            await e.save()
        return await super().save()

    async def archive(self) -> ArchivedTransaction:
        """Archive the Transaction. If it has already been archived,
            return the existing ArchivedTransaction.
        """
        try:
            return await ArchivedTransaction.insert({**self.data})
        except Exception as e:
            return await ArchivedTransaction.find(self.id)
